@@ -1,6 +1,6 @@
 /************************************************************************
 **
-**  Copyright (C) 2015-2020  Kevin B. Hendricks Stratford, ON, Canada 
+**  Copyright (C) 2015-2021  Kevin B. Hendricks Stratford, ON, Canada 
 **  Copyright (C) 2009-2011  Strahinja Markovic  <strahinja.markovic@gmail.com>
 **
 **  This file is part of Sigil.
@@ -20,17 +20,18 @@
 **
 *************************************************************************/
 
-#include <QtCore/QtCore>
-#include <QtCore/QFileInfo>
-#include <QtCore/QFutureSynchronizer>
+#include <QtCore>
+#include <QFileInfo>
+#include <QFutureSynchronizer>
 #include <QtConcurrent/QtConcurrent>
-#include <QtWidgets/QApplication>
-#include <QtWidgets/QProgressDialog>
+#include <QApplication>
+#include <QProgressDialog>
 
 #include "BookManipulation/Book.h"
 #include "BookManipulation/CleanSource.h"
 #include "BookManipulation/FolderKeeper.h"
-#include "Misc/GumboInterface.h"
+#include "Parsers/GumboInterface.h"
+#include "Parsers/CSSToolbox.h"
 #include "Misc/TempFolder.h"
 #include "Misc/Utility.h"
 #include "Misc/HTMLSpellCheck.h"
@@ -39,6 +40,7 @@
 #include "ResourceObjects/HTMLResource.h"
 #include "ResourceObjects/NCXResource.h"
 #include "ResourceObjects/OPFResource.h"
+#include "ResourceObjects/MiscTextResource.h"
 #include "sigil_constants.h"
 #include "SourceUpdates/AnchorUpdates.h"
 #include "SourceUpdates/PerformHTMLUpdates.h"
@@ -46,6 +48,7 @@
 #include "Misc/SettingsStore.h"
 
 static const QString FIRST_CSS_NAME   = "Style0001.css";
+static const QString FIRST_JS_NAME    = "Script0001.js";
 static const QString FIRST_SVG_NAME   = "Image0001.svg";
 static const QString PLACEHOLDER_TEXT = "PLACEHOLDER";
 static const QString EMPTY_HTML_FILE  = "<?xml version=\"1.0\" encoding=\"utf-8\"?>\n"
@@ -87,14 +90,14 @@ const QString EMPTY_NAV_FILE_START =
     "<html xmlns=\"http://www.w3.org/1999/xhtml\" xmlns:epub=\"http://www.idpf.org/2007/ops\" "
     "lang=\"%1\" xml:lang=\"%2\">\n"
     "<head>\n"
-    "  <title></title>\n"
+    "  <title>ePub NAV</title>\n"
     "  <meta charset=\"utf-8\" />\n"
     "  <link href=\"%3\" rel=\"stylesheet\" type=\"text/css\"/>"
     "</head>\n"
     "<body epub:type=\"frontmatter\">\n";
 
 const QString EMPTY_NAV_FILE_TOC = 
-    "  <nav epub:type=\"toc\" id=\"toc\">\n"
+    "  <nav epub:type=\"toc\" id=\"toc\" role=\"doc-toc\">\n"
     "    <h1>%1</h1>\n"
     "    <ol>\n"
     "      <li>\n"
@@ -333,9 +336,9 @@ HTMLResource *Book::CreateNewHTMLFile(const QString &folder_path)
     Utility::WriteUnicodeTextFile(PLACEHOLDER_TEXT, fullfilepath);
     Resource * resource = m_Mainfolder->AddContentFileToFolder(fullfilepath, 
                                                                true, 
-							       QString("application/xhtml+xml"),
-							       QString(),
-							       folder_path);
+                                                               QString("application/xhtml+xml"),
+                                                               QString(),
+                                                               folder_path);
     HTMLResource *html_resource = qobject_cast<HTMLResource *>(resource);
     SetModified(true);
     return html_resource;
@@ -357,16 +360,16 @@ HTMLResource *Book::CreateEmptyHTMLFile(const QString &folderpath)
 
 
 HTMLResource *Book::CreateEmptyNavFile(bool update_opf,
-				       const QString &folderpath,
-				       const QString &navname,
-				       const QString &first_textdir)
+                                       const QString &folderpath,
+                                       const QString &navname,
+                                       const QString &first_textdir)
 {
     bool found_css = false;
     Resource * styleresource = NULL;
     QList<Resource*> resources = GetFolderKeeper()->GetResourceTypeAsGenericList<CSSResource>(false);
     foreach(Resource *resource, resources) {
         if (resource->Filename() == SGC_NAV_CSS_FILENAME) {
-	    styleresource = resource;
+            styleresource = resource;
             found_css = true;
             break;
         }
@@ -381,8 +384,8 @@ HTMLResource *Book::CreateEmptyNavFile(bool update_opf,
             Utility::WriteUnicodeTextFile(SGC_NAV_CSS_FILE, css_path);
         }
         styleresource = m_Mainfolder->AddContentFileToFolder(css_path, 
-							     update_opf, 
-							     "text/css");
+                                                             update_opf, 
+                                                             "text/css");
         CSSResource *css_resource = qobject_cast<CSSResource *> (styleresource);
         // Need to make sure InitialLoad is done in newly added css resource object to prevent
         // blank css issues after a save to disk
@@ -393,10 +396,10 @@ HTMLResource *Book::CreateEmptyNavFile(bool update_opf,
     QString fullfilepath = tempfolder.GetPath() + "/" + navname;
     Utility::WriteUnicodeTextFile(PLACEHOLDER_TEXT, fullfilepath);
     Resource * resource = m_Mainfolder->AddContentFileToFolder(fullfilepath,
-							       update_opf, 
-							       "application/xhtml+xml",
-							       QString(),
-							       folderpath);
+                                                               update_opf, 
+                                                               "application/xhtml+xml",
+                                                               QString(),
+                                                               folderpath);
 
     // get the informtion we need to correctly fill the template
     QString navbookpath = resource->GetRelativePath();
@@ -425,7 +428,9 @@ HTMLResource *Book::CreateEmptyNavFile(bool update_opf,
     return html_resource;
 }
 
-
+#if 0
+// Note: Broken -  CreateNewHTMLFile already appends to the end of the spine and if needed
+// You can move using the OPFResource to change Reading Order much faster
 HTMLResource *Book::CreateEmptyHTMLFile(HTMLResource *resource, const QString &folderpath)
 {
     HTMLResource *new_resource = CreateNewHTMLFile(folderpath);
@@ -448,6 +453,7 @@ HTMLResource *Book::CreateEmptyHTMLFile(HTMLResource *resource, const QString &f
     SetModified(true);
     return new_resource;
 }
+#endif
 
 
 void Book::MoveResourceAfter(HTMLResource *from_resource, HTMLResource *to_resource)
@@ -456,18 +462,11 @@ void Book::MoveResourceAfter(HTMLResource *from_resource, HTMLResource *to_resou
         return;
     }
 
-    QList<HTMLResource *> html_resources = m_Mainfolder->GetResourceTypeList<HTMLResource>(true);
-    int to_after_reading_order = GetOPF()->GetReadingOrder(to_resource) + 1;
-    int from_reading_order = GetOPF()->GetReadingOrder(from_resource) ;
-
-    if (to_after_reading_order > 0) {
-        html_resources.move(from_reading_order, to_after_reading_order);
-        GetOPF()->UpdateSpineOrder(html_resources);
-    }
-
+    GetOPF()->MoveReadingOrder(from_resource, to_resource);
     SetModified(true);
 }
 
+
 HTMLResource *Book::CreateHTMLCoverFile(QString text)
 {
     HTMLResource *html_resource = CreateNewHTMLFile();
@@ -516,15 +515,30 @@ CSSResource *Book::CreateEmptyCSSFile(const QString &folderpath)
     QString fullfilepath = tempfolder.GetPath() + "/" + m_Mainfolder->GetUniqueFilenameVersion(FIRST_CSS_NAME);
     Utility::WriteUnicodeTextFile("", fullfilepath);
     Resource * resource = m_Mainfolder->AddContentFileToFolder(fullfilepath,
-							       true,
-							       "text/css",
-							       QString(),
-							       folderpath);
+                                                               true,
+                                                               "text/css",
+                                                               QString(),
+                                                               folderpath);
     CSSResource *css_resource = qobject_cast<CSSResource *>(resource);
     SetModified(true);
     return css_resource;
 }
 
+MiscTextResource *Book::CreateEmptyJSFile(const QString &folderpath)
+{
+    TempFolder tempfolder;
+    QString fullfilepath = tempfolder.GetPath() + "/" + m_Mainfolder->GetUniqueFilenameVersion(FIRST_JS_NAME);
+    Utility::WriteUnicodeTextFile("", fullfilepath);
+    Resource * resource = m_Mainfolder->AddContentFileToFolder(fullfilepath,
+                                                               true,
+                                                               "application/javascript",
+                                                               QString(),
+                                                               folderpath);
+    MiscTextResource *js_resource = qobject_cast<MiscTextResource *>(resource);
+    SetModified(true);
+    return js_resource;
+}
+
 
 SVGResource *Book::CreateEmptySVGFile(const QString& folderpath)
 {
@@ -630,12 +644,15 @@ void Book::CreateNewSections(const QStringList &new_sections, HTMLResource *orig
         sectionInfo.new_file_prefix = new_file_prefix;
         sectionInfo.file_suffix = i;
         sectionInfo.file_extension = file_extension;
-	sectionInfo.folder_path = folder_path;
+        sectionInfo.folder_path = folder_path;
         sync.addFuture(
-            QtConcurrent::run(
-                this,
-                &Book::CreateOneNewSection,
-                sectionInfo));
+#if QT_VERSION < QT_VERSION_CHECK(6, 0, 0)                      
+            QtConcurrent::run(this, &Book::CreateANewSection, sectionInfo)
+#else
+            QtConcurrent::run(&Book::CreateANewSection, this, sectionInfo)
+            //QtConcurrent::run(this, [this, sectionInfo] { &Book::CreateOneNewSection(sectionInfo);
+#endif
+                      );
     }
 
     sync.waitForFinished();
@@ -703,6 +720,43 @@ bool Book::IsDataOnDiskWellFormed(HTMLResource *html_resource)
     return error.line == -1;
 }
 
+
+bool Book::RenameClassInHTML(const QString css_bookpath, const QString oldname, const QString newname)
+{
+    const QList<HTMLResource *> html_resources = m_Mainfolder->GetResourceTypeList<HTMLResource>(false);
+    QFuture< bool > rfuture;
+    rfuture = QtConcurrent::mapped(html_resources, std::bind(Book::RenameClassInHTMLFileMapped,
+                                                             std::placeholders::_1,
+                                                             css_bookpath,
+                                                             oldname,
+                                                             newname));
+    bool result = true;
+    for (int i = 0; i < rfuture.results().count(); i++) {
+        result = result && rfuture.resultAt(i);
+    }
+    return result;
+}
+
+
+bool Book::RenameClassInHTMLFileMapped(HTMLResource* html_resource,
+                                       const QString css_bookpath,
+                                       const QString oldname,
+                                       const QString newname)
+{
+    QStringList linked_stylesheets = html_resource->GetLinkedStylesheets();
+    QString xhtmltext = html_resource->GetText();
+    if (linked_stylesheets.contains(css_bookpath)) {
+        QWriteLocker locker(&html_resource->GetLock());
+        CSSToolbox tb;
+        QString newtext = tb.rename_class_in_text(oldname, newname, xhtmltext);
+        if (xhtmltext != newtext) {
+            html_resource->SetText(newtext);
+        }
+    }
+    return true;
+}
+
+
 void Book::ReformatAllHTML(bool to_valid)
 {
     QApplication::setOverrideCursor(Qt::WaitCursor);
@@ -781,9 +835,9 @@ std::tuple<QString, QStringList> Book::GetStyleUrlsInHTMLFileMapped(HTMLResource
     foreach (QString url, urllist) {
         QRegularExpressionMatch match = url_file_search.match(url);
         if (match.hasMatch()) {
-	    QString ahref = match.captured(1);
+            QString ahref = match.captured(1);
             if (ahref.indexOf(":") == -1) {
-	        bookpaths << Utility::buildBookPath(ahref, startdir);
+                bookpaths << Utility::buildBookPath(ahref, startdir);
             }
         }
     }
@@ -893,7 +947,7 @@ QHash< QString, std::pair<int,int> > Book::GetSpellWordCountsInHTMLFiles()
     QFuture<std::tuple<QString, std::pair<int,int> > > future = QtConcurrent::mapped(html_resources, GetWordCountsInHTMLFileMapped);
     for (int i = 0; i < future.results().count(); i++) {
         QString bookpath;
-	std::pair<int, int> word_counts;
+        std::pair<int, int> word_counts;
         std::tie(bookpath, word_counts) = future.resultAt(i);
         words_in_html[bookpath] = word_counts;
     }
@@ -990,7 +1044,7 @@ std::tuple<QString, QStringList> Book::GetMediaInHTMLFileMapped(HTMLResource *ht
     foreach(QString ahref, media_hrefs) {
         if (ahref.indexOf(":") == -1) {
             media_bookpaths << Utility::buildBookPath(ahref, startdir);
-	}
+        }
     }
     return std::make_tuple(html_bookpath, media_bookpaths);
 }
@@ -1003,7 +1057,7 @@ std::tuple<QString, QStringList> Book::GetImagesInHTMLFileMapped(HTMLResource *h
     QStringList image_bookpaths;
     foreach(QString ahref, image_hrefs) {
         if (ahref.indexOf(":") == -1) {
-	    image_bookpaths << Utility::buildBookPath(ahref, startdir);
+            image_bookpaths << Utility::buildBookPath(ahref, startdir);
         }
     }
     return std::make_tuple(html_bookpath, image_bookpaths);
@@ -1017,7 +1071,7 @@ std::tuple<QString, QStringList> Book::GetVideoInHTMLFileMapped(HTMLResource *ht
     QStringList video_bookpaths;
     foreach(QString ahref, video_hrefs) {
         if (ahref.indexOf(":") == -1) {
-	    video_bookpaths << Utility::buildBookPath(ahref, startdir);
+            video_bookpaths << Utility::buildBookPath(ahref, startdir);
         }
     }
     return std::make_tuple(html_bookpath, video_bookpaths);
@@ -1031,12 +1085,24 @@ std::tuple<QString, QStringList> Book::GetAudioInHTMLFileMapped(HTMLResource *ht
     QStringList audio_bookpaths;
     foreach(QString ahref, audio_hrefs) {
         if (ahref.indexOf(":") == -1) {
-	    audio_bookpaths << Utility::buildBookPath(ahref, startdir);
+            audio_bookpaths << Utility::buildBookPath(ahref, startdir);
         }
     }
     return std::make_tuple(html_bookpath, audio_bookpaths);
 }
 
+bool Book::CheckHTMLFilesForWellFormedness(const QList<HTMLResource*> html_resources)
+{
+    bool wellformed = true;
+    QFuture< std::pair<HTMLResource*, bool> > well_future;
+    well_future = QtConcurrent::mapped(html_resources, ResourceWellFormedMap);
+    for (int i = 0; i < well_future.results().count(); i++) {
+        std::pair<HTMLResource*, bool> res = well_future.resultAt(i);
+        wellformed = wellformed && res.second;
+    }
+    return wellformed;
+}
+
 QList<HTMLResource *> Book::GetNonWellFormedHTMLFiles()
 {
     QList<HTMLResource *> malformed_resources;
@@ -1044,7 +1110,7 @@ QList<HTMLResource *> Book::GetNonWellFormedHTMLFiles()
     QFuture< std::pair<HTMLResource*, bool> > well_future;
     well_future = QtConcurrent::mapped(html_resources, ResourceWellFormedMap);
     for (int i = 0; i < well_future.results().count(); i++) {
-	std::pair<HTMLResource*, bool> res = well_future.resultAt(i);
+        std::pair<HTMLResource*, bool> res = well_future.resultAt(i);
         if (!res.second) malformed_resources << res.first;
     }
     return malformed_resources;
@@ -1066,15 +1132,20 @@ QSet<QString> Book::GetWordsInHTMLFiles()
     default_lang.replace('_','-');
     const QList<HTMLResource *> html_resources = m_Mainfolder->GetResourceTypeList<HTMLResource>(false);
     QFuture<QStringList> future = QtConcurrent::mapped(html_resources, std::bind(GetWordsInHTMLFileMapped,
-										 std::placeholders::_1,
-										 default_lang));
+                                                                                 std::placeholders::_1,
+                                                                                 default_lang));
 
     for (int i = 0; i < future.results().count(); i++) {
         QStringList result = future.resultAt(i);
         all_words.append(result);
     }
 
-    return all_words.toSet();  // Qt 5.15:  QSet<QString>(all_words.begin(), all_words.end());
+#if QT_VERSION < QT_VERSION_CHECK(5, 15, 0)
+    return all_words.toSet();
+#else
+    QSet<QString> allwordset(all_words.begin(), all_words.end());
+    return allwordset;
+#endif
 }
 
 QStringList Book::GetWordsInHTMLFileMapped(HTMLResource *html_resource, const QString& default_lang)
@@ -1092,8 +1163,8 @@ QHash<QString, int> Book::GetUniqueWordsInHTMLFiles()
     QHash<QString, int> all_words;
     const QList<HTMLResource *> html_resources = m_Mainfolder->GetResourceTypeList<HTMLResource>(false);
     QFuture<QStringList> future = QtConcurrent::mapped(html_resources, std::bind(GetWordsInHTMLFileMapped,
-										 std::placeholders::_1,
-										 default_lang));
+                                                                                 std::placeholders::_1,
+                                                                                 default_lang));
 
     for (int i = 0; i < future.results().count(); i++) {
         QStringList result = future.resultAt(i);
@@ -1154,6 +1225,50 @@ QStringList Book::GetStylesheetsInHTMLFile(HTMLResource *html_resource)
     return results;
 }
 
+// employed to perform update of local merged links and to extract the updated body
+QPair<QString, QString> Book::UpdateAndExtractBodyInOneFile(Resource * resource, const QStringList & merged_bookpaths)
+{
+    QPair<QString, QString> res;
+    HTMLResource *htmlresource = qobject_cast<HTMLResource *>(resource);
+    if (!htmlresource) return res;
+
+    QReadLocker locker(&htmlresource->GetLock());
+    QString startdir = htmlresource->GetFolder();
+    QString version = htmlresource->GetEpubVersion();
+    QString bookpath = htmlresource->GetRelativePath();
+    GumboInterface gi = GumboInterface(htmlresource->GetText(), version);
+    gi.parse();
+    const QList<GumboNode*> anchor_nodes = gi.get_all_nodes_with_tag(GUMBO_TAG_A);
+
+    for (int i = 0; i < anchor_nodes.length(); ++i) {
+        GumboNode* node = anchor_nodes.at(i);
+        GumboAttribute* attr = gumbo_get_attribute(&node->v.element.attributes, "href");
+
+        // We find the hrefs that are relative and contain an href. If they
+        // point into one of the merged resources, make it a local link
+        if (attr) {
+            QString href = QString::fromUtf8(attr->value);
+            if (href.indexOf(':') == -1) {
+                std::pair<QString, QString> parts = Utility::parseRelativeHREF(href);
+                QString target_bookpath = Utility::buildBookPath(parts.first, startdir);
+                if (merged_bookpaths.contains(target_bookpath)) {
+                    // remove the file path as it is now local
+                    if (parts.second.isEmpty()) parts.second = "#";
+                    QString attribute_value = Utility::buildRelativeHREF("", parts.second);
+                    gumbo_attribute_set_value(attr, attribute_value.toUtf8().constData());
+                }
+            }
+        }
+    }
+    res.first = bookpath;
+    res.second = gi.get_body_contents();
+    return res;
+}
+
+
+// first resource in list is the sink resource to merge into
+// It is the user's responsibility to ensure that all ids used across the two merged files are unique.
+// Reconcile all references to the files that were merged.
 Resource *Book::MergeResources(QList<Resource *> resources)
 {
 
@@ -1164,75 +1279,101 @@ Resource *Book::MergeResources(QList<Resource *> resources)
         return nav_resource;
     }
 
-    QProgressDialog progress(QObject::tr("Merging Files.."), 0, 0, resources.count(), QApplication::activeWindow());
-    progress.setMinimumDuration(PROGRESS_BAR_MINIMUM_DURATION);
-    int progress_value = 0;
-    Resource *sink_resource = resources.takeFirst();
+    // nothing to merge
+    if (resources.size() < 2) return NULL;
+
+    // First Create a set of Ids used in the files to be merged
+    QHash<QString,QStringList> BookPathIds = GetIdsInHTMLFiles();
+    QSet<QString> UsedIds;
+    foreach(Resource * resource, resources) {
+        QString bookpath = resource->GetRelativePath();
+        QStringList ids=BookPathIds.value(bookpath, QStringList());
+        foreach(QString id, ids) {
+            UsedIds.insert(id);
+        }
+    }
+
+    // Create the list of bookpaths being merged into in order
+    QList<QString> merged_bookpaths;
+    foreach(Resource * resource, resources) {
+        merged_bookpaths << resource->GetRelativePath();
+    }
+
+    Resource *sink_resource = resources.at(0);
     HTMLResource *sink_html_resource = qobject_cast<HTMLResource *>(sink_resource);
 
-    if (!IsDataWellFormed(sink_html_resource)) {
-        return sink_resource;
+    const QList<QPair<QString, QString>> &bodies = QtConcurrent::blockingMapped(resources, 
+                                                                                std::bind(UpdateAndExtractBodyInOneFile, 
+                                                                                          std::placeholders::_1,
+                                                                                          merged_bookpaths));
+    // collect the outputs from the many threads
+    QHash<QString, QString> updated_bodies;
+    for (int i = 0; i < bodies.count(); ++i) {
+        QPair<QString, QString> body = bodies.at(i);
+        updated_bodies.insert(body.first, body.second);
     }
 
     QStringList new_bodies;
-    QList<QString> merged_bookpaths;
-    QString version = sink_html_resource->GetEpubVersion();
-    {
-        GumboInterface gi = GumboInterface(sink_html_resource->GetText(), version);
-        new_bodies << gi.get_body_contents();
-        Resource *failed_resource = NULL;
-        
-        // Now iterate across all the other resources merging them into this resource
-        foreach(Resource *source_resource, resources) {
+    QHash<QString,QString> section_id_map;
 
-            // Set progress value and ensure dialog has time to display when doing extensive updates
-            progress.setValue(progress_value++);
-            qApp->processEvents(QEventLoop::ExcludeUserInputEvents);
+    // remove everything after the body tag in the sink resource
+    QString text = sink_html_resource->GetText();
+    QRegularExpression body_search(BODY_START, QRegularExpression::CaseInsensitiveOption);
+    QRegularExpressionMatch body_search_mo = body_search.match(text);
+    int body_tag_end   = body_search_mo.capturedStart() + body_search_mo.capturedLength();
+    new_bodies << Utility::Substring(0, body_tag_end, text);
 
-            HTMLResource *source_html_resource = qobject_cast<HTMLResource *>(source_resource);
-            if (!IsDataWellFormed(source_html_resource)) {
-                failed_resource = source_resource;
-                break;
-            }
-
-            // Get the html document for this source resource.
-            GumboInterface ngi = GumboInterface(source_html_resource->GetText(), version);
-            new_bodies << ngi.get_body_contents();
-            merged_bookpaths.append(source_resource->GetRelativePath());
+    // build up the new merged file and set it into the sink resource
+    int i = 0;
+    foreach(QString bookpath, merged_bookpaths) {
+        // inject anchor tag to start each merged section and record it after the sink resource
+        if (i == 0) {
+            new_bodies.append(updated_bodies[bookpath]);
+        } else {
+            QString section_id = Utility::GenerateUniqueId("section", UsedIds);
+            UsedIds.insert(section_id);
+            new_bodies.append("  <a id=\"" + section_id + "\"></a>\n" + updated_bodies[bookpath]);
+            section_id_map[bookpath] = section_id;
         }
+        i++;
+    }
+    new_bodies.append("</body>\n</html>");
+    QString new_source = new_bodies.join("");
+    sink_html_resource->SetText(new_source);
 
-        if (failed_resource != NULL) {
-            // Abort the merge process. We will discard whatever we had merged so far
-            return failed_resource;
-        }
-        QString new_body = new_bodies.join("");
-        QString new_source = gi.perform_body_updates(new_body);
-        // Now all fragments have been merged into this sink document, serialize and store it.
-        sink_html_resource->SetText(new_source);
-        // Now safe to do the delete
-        foreach(Resource *source_resource, resources) {
-            // Need to alert FolderKeeper that these are going away to properly update its
-            // m_Resources hash to prevent stale values from deleted resources hanging around
-            m_Mainfolder->RemoveResource(source_resource);
-            source_resource->Delete();
+    // Anchor Updates should handle updating all links in the nav properly
+    // But if this is an epub2, then we must update the guide entries as well
+    QString version = sink_resource->GetEpubVersion(); 
+    if (version.startsWith("2")) {
+        GetOPF()->UpdateGuideAfterMerge(resources, section_id_map);
+    }
+
+    // Update all anchors links from outside the merged set into it
+    QList<HTMLResource*> html_resources;
+    foreach(HTMLResource* hres, m_Mainfolder->GetResourceTypeList<HTMLResource>(true)) {
+        if (!merged_bookpaths.contains(hres->GetRelativePath())) {
+            html_resources << hres;
         }
     }
-    progress.setValue(progress_value++);
-    qApp->processEvents(QEventLoop::ExcludeUserInputEvents);
-    // It is the user's responsibility to ensure that all ids used across the two merged files are unique.
-    // Reconcile all references to the files that were merged.
-    QList<HTMLResource *> html_resources = m_Mainfolder->GetResourceTypeList<HTMLResource>(true);
-    AnchorUpdates::UpdateAllAnchors(html_resources, merged_bookpaths, sink_html_resource);
+    AnchorUpdates::UpdateAllAnchors(html_resources, merged_bookpaths, sink_html_resource, section_id_map);
+
+    // Update any NCX anchors into the merged set
     NCXResource * ncx_resource = GetNCX();
     if (ncx_resource) {
         AnchorUpdates::UpdateTOCEntriesAfterMerge(ncx_resource, 
-						  sink_html_resource->GetRelativePath(),
-						  merged_bookpaths);
+                                                  sink_html_resource->GetRelativePath(),
+                                                  merged_bookpaths);
     }
+
+    // finally delete the merged resources except for the sink
+    resources.removeOne(sink_resource);
+    m_Mainfolder->BulkRemoveResources(resources);
+
     SetModified(true);
     return NULL;
 }
 
+
 QList <Resource *> Book::GetAllResources()
 {
     return m_Mainfolder->GetResourceList();
@@ -1302,26 +1443,26 @@ std::tuple<bool, QString, QString> Book::HasUndefinedURLFragments()
 
     foreach(HTMLResource *html_resource, html_resources) {
         QString bookpath = html_resource->GetRelativePath();
-	QString htmldir = html_resource->GetFolder();
+        QString htmldir = html_resource->GetFolder();
         foreach(QString ahref, links[bookpath]) {
-	    // each links is relative to the bookpath resource and raw
+            // each links is relative to the bookpath resource and raw
             // convert this href to a bookpath href
-	    std::pair<QString,QString> hrefparts = Utility::parseRelativeHREF(ahref);
-	    QString attpath = hrefparts.first;
-	    QString dest_bookpath;
-	    if (attpath.isEmpty()) {
-	        dest_bookpath = bookpath;
-	    } else {
-	        dest_bookpath = Utility::buildBookPath(attpath, htmldir);
-	    }
-	    QString dest_id = hrefparts.second;
-	    if (dest_id.startsWith("#")) dest_id = dest_id.mid(1,-1);
+            std::pair<QString,QString> hrefparts = Utility::parseRelativeHREF(ahref);
+            QString attpath = hrefparts.first;
+            QString dest_bookpath;
+            if (attpath.isEmpty()) {
+                dest_bookpath = bookpath;
+            } else {
+                dest_bookpath = Utility::buildBookPath(attpath, htmldir);
+            }
+            QString dest_id = hrefparts.second;
+            if (dest_id.startsWith("#")) dest_id = dest_id.mid(1,-1);
 
-	    if (!dest_id.isEmpty()) {
-	        if (html_bookpaths.contains(dest_bookpath) && !all_ids[dest_bookpath].contains(dest_id)) {
-		    return std::make_tuple(true, ahref, bookpath);
-	        }
-	    }
+            if (!dest_id.isEmpty()) {
+                if (html_bookpaths.contains(dest_bookpath) && !all_ids[dest_bookpath].contains(dest_id)) {
+                    return std::make_tuple(true, ahref, bookpath);
+                }
+            }
         }
     }
     return std::make_tuple(false, QString(), QString());
@@ -1357,7 +1498,7 @@ void Book::SaveOneResourceToDisk(Resource *resource)
 }
 
 
-Book::NewSectionResult Book::CreateOneNewSection(NewSection section_info)
+Book::NewSectionResult Book::CreateANewSection(NewSection section_info)
 {
     return CreateOneNewSection(section_info, QHash<QString, QString>());
 }
@@ -1370,10 +1511,10 @@ Book::NewSectionResult Book::CreateOneNewSection(NewSection section_info,
     QString fullfilepath = section_info.temp_folder_path + "/" + filename;
     Utility::WriteUnicodeTextFile("PLACEHOLDER", fullfilepath);
     Resource * resource = m_Mainfolder->AddContentFileToFolder(fullfilepath, 
-							       true, 
-							       QString(), 
-							       QString(), 
-							       section_info.folder_path);
+                                                               true, 
+                                                               QString(), 
+                                                               QString(), 
+                                                               section_info.folder_path);
     HTMLResource *html_resource = qobject_cast<HTMLResource *>(resource);
     Q_ASSERT(html_resource);
     QString version = html_resource->GetEpubVersion();
@@ -1382,11 +1523,11 @@ Book::NewSectionResult Book::CreateOneNewSection(NewSection section_info,
         html_resource->SetText(CleanSource::Mend(section_info.source, version));
     } else {
         QString currentpath = html_resource->GetCurrentBookRelPath();
-	QString newbookpath = html_resource->GetRelativePath();
+        QString newbookpath = html_resource->GetRelativePath();
         html_resource->SetText(PerformHTMLUpdates(CleanSource::Mend(section_info.source, version),
-					     newbookpath,
-                                             html_updates, QHash<QString, QString>(), 
-                                             currentpath, version)() );
+                                                  newbookpath,
+                                                  html_updates, QHash<QString, QString>(), 
+                                                  currentpath, version)() );
         html_resource->SetCurrentBookRelPath("");
     }
 
