@@ -1,6 +1,6 @@
 /************************************************************************
 **
-**  Copyright (C) 2015-2020 Kevin B. Hendricks, Stratford Ontario Canada
+**  Copyright (C) 2015-2022 Kevin B. Hendricks, Stratford Ontario Canada
 **  Copyright (C) 2009-2011 Strahinja Markovic  <strahinja.markovic@gmail.com>
 **
 **  This file is part of Sigil.
@@ -32,13 +32,13 @@
 #include <iowin32.h>
 #endif
 
-#include <QtCore/QDateTime>
-#include <QtCore/QDir>
-#include <QtCore/QDirIterator>
-#include <QtCore/QFile>
-#include <QtCore/QFileInfo>
-#include <QtCore/QTemporaryFile>
-#include <QtCore/QTextStream>
+#include <QDateTime>
+#include <QDir>
+#include <QDirIterator>
+#include <QFile>
+#include <QFileInfo>
+#include <QTemporaryFile>
+#include <QTextStream>
 
 #include "BookManipulation/CleanSource.h"
 #include "BookManipulation/FolderKeeper.h"
@@ -48,6 +48,7 @@
 #include "Misc/Utility.h"
 #include "Misc/TempFolder.h"
 #include "Misc/FontObfuscation.h"
+#include "ResourceObjects/Resource.h"
 #include "ResourceObjects/FontResource.h"
 #include "sigil_constants.h"
 #include "sigil_exception.h"
@@ -91,7 +92,9 @@ void ExportEPUB::WriteBook()
         m_Book->GetOPF()->EnsureUUIDIdentifierPresent();
     }
 
-    m_Book->GetOPF()->AddSigilVersionMeta();
+    if (Utility::GetEnvironmentVar("SIGIL_DISABLE_VERSION_META").isEmpty()) {
+        m_Book->GetOPF()->AddSigilVersionMeta();
+    }
     m_Book->GetOPF()->AddModificationDateMeta();
     m_Book->SaveAllResourcesToDisk();
     TempFolder tempfolder;
@@ -104,7 +107,6 @@ void ExportEPUB::WriteBook()
     SaveFolderAsEpubToLocation(tempfolder.GetPath(), m_FullFilePath);
 }
 
-
 // Creates the publication from the Book
 // (creates XHTML, CSS, OPF, NCX files etc.)
 void ExportEPUB::CreatePublication(const QString &fullfolderpath)
@@ -120,6 +122,7 @@ void ExportEPUB::SaveFolderAsEpubToLocation(const QString &fullfolderpath, const
 {
     QString tempFile = fullfolderpath + "-tmp.epub";
     QDateTime timeNow = QDateTime::currentDateTime();
+    QString modified_now = timeNow.toString("yyyy-MM-dd hh:mm:ss");
     zip_fileinfo fileInfo;
 #ifdef Q_OS_WIN32
     zlib_filefunc64_def ffunc;
@@ -134,14 +137,14 @@ void ExportEPUB::SaveFolderAsEpubToLocation(const QString &fullfolderpath, const
     }
 
     memset(&fileInfo, 0, sizeof(fileInfo));
-    fileInfo.tmz_date.tm_sec = timeNow.time().second();
-    fileInfo.tmz_date.tm_min = timeNow.time().minute();
+    fileInfo.tmz_date.tm_sec  = timeNow.time().second();
+    fileInfo.tmz_date.tm_min  = timeNow.time().minute();
     fileInfo.tmz_date.tm_hour = timeNow.time().hour();
     fileInfo.tmz_date.tm_mday = timeNow.date().day();
-    fileInfo.tmz_date.tm_mon = timeNow.date().month() - 1;
+    fileInfo.tmz_date.tm_mon  = timeNow.date().month() - 1;
     fileInfo.tmz_date.tm_year = timeNow.date().year();
 
-    // Write the mimetype. This must be uncompressed and the first entry in the archive.
+     // Write the mimetype. This must be uncompressed and the first entry in the archive.
     if (zipOpenNewFileInZip64(zfile, "mimetype", &fileInfo, NULL, 0, NULL, 0, NULL, Z_NO_COMPRESSION, 0, 0) != ZIP_OK) {
         zipClose(zfile, NULL);
         QFile::remove(tempFile);
@@ -167,6 +170,37 @@ void ExportEPUB::SaveFolderAsEpubToLocation(const QString &fullfolderpath, const
             relpath = relpath.remove(0, 1);
         }
 
+        // do not double add the mimetype file
+        if (relpath == "mimetype") continue;
+
+        QFileInfo tfile(it.filePath());
+
+        // Set the proper zip file info if possible
+        QString amodified = modified_now;
+        size_t afilesize = tfile.size();
+        QString afilecrc = Utility::FileCRC32(it.filePath());
+        Resource* resource = m_Book->GetFolderKeeper()->GetResourceByBookPathNoThrow(relpath);
+        if (resource) {
+            QString savedcrc  = resource->GetSavedCRC32();
+            QString saveddate = resource->GetSavedDate();
+            size_t savedsize = resource->GetSavedSize();
+            if ( (savedsize == afilesize) && (savedcrc == afilecrc) ) {
+                amodified = saveddate;
+            } else {
+                resource->SetSavedDate(amodified);
+                resource->SetSavedSize(afilesize);
+                resource->SetSavedCRC32(afilecrc);
+            }
+        }
+        QDateTime moddate = QDateTime::fromString(amodified, "yyyy-MM-dd hh:mm:ss");
+        memset(&fileInfo, 0, sizeof(fileInfo));
+        fileInfo.tmz_date.tm_sec  = moddate.time().second();
+        fileInfo.tmz_date.tm_min  = moddate.time().minute();
+        fileInfo.tmz_date.tm_hour = moddate.time().hour();
+        fileInfo.tmz_date.tm_mday = moddate.date().day();
+        fileInfo.tmz_date.tm_mon  = moddate.date().month() - 1;
+        fileInfo.tmz_date.tm_year = moddate.date().year();
+
         // Add the file entry to the archive.
         // We should check the uncompressed file size. If it's over >= 0xffffffff the last parameter (zip64) should be 1.
         if (zipOpenNewFileInZip4_64(zfile, relpath.toUtf8().constData(), &fileInfo, NULL, 0, NULL, 0, NULL, Z_DEFLATED, 8, 0, 15, 8, Z_DEFAULT_STRATEGY, NULL, 0, 0x0b00, 1<<11, 0) != ZIP_OK) {
@@ -175,8 +209,7 @@ void ExportEPUB::SaveFolderAsEpubToLocation(const QString &fullfolderpath, const
             throw(CannotStoreFile(relpath.toStdString()));
         }
 
-        // Open the file on disk. We will read this and write what we read into
-        // the archive.
+        // the file on disk to write
         QFile dfile(it.filePath());
 
         if (!dfile.open(QIODevice::ReadOnly)) {
