@@ -25,6 +25,17 @@
 #include <QStringList>
 #include "sigil_constants.h"
 
+const QString CLIPS_SETTINGS_FILE = "sigil_clips.ini";
+const QString INDEX_SETTINGS_FILE = "sigil_index.ini";
+const QString SEARCHES_SETTINGS_FILE = "sigil_searches.ini";
+const QString SEARCHES_V2_SETTINGS_FILE = "sigil_searches_v2.ini";
+const QString SIGIL_SETTINGS_FILE = "sigil.ini";
+
+const QString CLIPS_V6_SETTINGS_FILE = "sigil_clips_v6.ini";
+const QString INDEX_V6_SETTINGS_FILE = "sigil_index_v6.ini";
+const QString SEARCHES_V6_SETTINGS_FILE = "sigil_searches_v6.ini";
+const QString SIGIL_V6_SETTINGS_FILE = "sigil_v6.ini";
+
 #if _WIN32
 #include <QProcessEnvironment>
 
